#pragma once
#include <cmath>
#include <boost\type_traits\remove_const.hpp>
#include <stdexcept>
#include <assert.h>
#include <Matrix3x3_fwd.h>

namespace Linear{

	

	static const double VEPSILON = 0.000001;

	template<typename NType>
	struct Vector{
	public:
		NType x;
		NType y;
		NType z;

		inline Vector():x(0),y(0),z(0){}

		template <typename OtherType>
		inline Vector(OtherType init_x, OtherType init_y, OtherType init_z)
			:x(init_x),y(init_y),z(init_z){
		}
		
		
		template <typename OtherType>
		inline Vector(const Vector<OtherType>& other)
			:x(other.x),y(other.y),z(other.z){
		}

		template <typename OtherType>
		inline Vector(const OtherType& other)
			:x(other.x),y(other.y),z(other.z){
		}

		template <typename OtherType>
		inline Vector& operator=(const Vector<OtherType>& other){
			x=other.x;
			y=other.y;
			z=other.z;
			return *this;
		}
		inline ~Vector(){}



		void Reset(){
			x = NType(0);
			y = NType(0);
			z = NType(0);
		}

		template <typename OtherType>
		void Reset(OtherType init_x, OtherType init_y, OtherType init_z){
			x=init_x;
			y=init_y;
			z=init_z;
		}

		inline operator bool() const {return x||y||z;}
		inline bool operator==(const Vector<NType>& other) const{
			return abs(other.x-x)<=NType(VEPSILON) && abs(other.y-y)<=NType(VEPSILON) && abs(other.z-z)<=NType(VEPSILON);
		}

		inline bool operator!=(const Vector<NType>& other) const{
			return abs(other.x-x)>NType(VEPSILON) || abs(other.y-y)>NType(VEPSILON) || abs(other.z-z)>NType(VEPSILON);
		}

		//ȉŗLo֐
		inline NType Length() const{ //xNg̑傫
			return sqrt(x*x+y*y+z*z);
		}

		inline Vector& Normalize(){	//xNg̐K
			NType m = sqrt( x*x + y*y + z*z );
			if(m <= NType(VEPSILON) )  throw std::runtime_error("Vector is too small.");
			auto m2 = m / NType(1);
			x /= m2;
			y /= m2;
			z /= m2;
	
			if( abs(x) < NType(VEPSILON)) x = NType(0);
			if( abs(y) < NType(VEPSILON)) y = NType(0);
			if( abs(z) < NType(VEPSILON)) z = NType(0);
			return *this;
		}

		template <typename OtherType>
		inline	Vector<NType>& operator+=(const Vector<OtherType>& v){
			x += v.x;
			y += v.y;
			z += v.z;
			return *this;
		}

		template <typename OtherType>
		inline	Vector<NType>& operator-=(const Vector<OtherType>& v){
			x -= v.x;
			y -= v.y;
			z -= v.z;
			return *this;
		}

		template <typename OtherType>
		inline Vector& operator*=(const OtherType s){
			x *= s;
			y *= s;
			z *= s;
			return *this;
		}

		template <typename OtherType>
		inline Vector& operator/=(const OtherType s){
			x /= s;
			y /= s;
			z /= s;
			return *this;
		}


		inline Vector<NType> operator-() const{
			return Vector<NType>(-x, -y, -z);
		}


		NType& operator[](int i){
			assert( 1 <= i && i <=3 );
			if( i == 1 ) 		return x;
			else if( i == 2 )	return y;
			else if( i == 3 )	return z;
			else				return z;
		}
		const NType& operator[](int i) const{
			assert( 1 <= i && i <=3 );
			if( i == 1 ) 		return x;
			else if( i == 2 )	return y;
			else if( i == 3 )	return z;
			else				return z;
		}




	};

	namespace Private{
		template <typename LHS, typename RHS >
		struct RCMul{
			typedef typename std::remove_const< decltype( LHS(0) * RHS(0) ) >::type type;
		};
		template <typename LHS, typename RHS >
		struct RCDev{
			typedef typename std::remove_const< decltype( LHS(0) / RHS(1) ) >::type type;
		};
		
		template <typename LHS, typename RHS>
		struct RCMul<Vector<LHS>, RHS>{
		};
 
		template <typename LHS, typename RHS>
		struct RCMul<LHS, Vector<RHS>>{
		}; 

		template <typename LHS, typename RHS>
		struct RCMul<Matrix3x3<LHS>, RHS>{
		};
 
		template <typename LHS, typename RHS>
		struct RCMul<LHS, Matrix3x3<RHS>>{
		}; 
	}

	//-------------------------------------------------------------------------------------------------------------------------------------

	template<typename NType>
	Vector< decltype( NType(0).GetValue() ) > GetValue(const Vector<NType>& lhs){
		return Vector< decltype(NType(0).GetValue()) >(lhs.x.GetValue(), lhs.y.GetValue(), lhs.z.GetValue());
	}



	//xNg̘aA
	template <typename NType>
	inline	Vector<NType> operator+(const Vector<NType>& lhs, const Vector<NType>& rhs){
		return Vector<NType>(lhs.x+rhs.x, lhs.y+rhs.y, lhs.z+rhs.z);
	}

	template <typename NType>
	inline	Vector<NType> operator-(const Vector<NType>& lhs, const Vector<NType>& rhs){
		return Vector<NType>(lhs.x-rhs.x, lhs.y-rhs.y, lhs.z-rhs.z);
	}
	
	 //
    template < typename NType1, typename NType2 >
    decltype( NType1(0) * NType2(0) )
    operator* ( const Vector<NType1>& lhs, const Vector<NType2>& rhs ){

        return lhs.x*rhs.x + lhs.y*rhs.y + lhs.z*rhs.z;
    
	}


	//O
	template <typename NType1, typename NType2>
	inline const Vector< typename Private::RCMul<NType1, NType2>::type >
	operator^(const Vector<NType1>& lhs, const Vector<NType2>& rhs){
		return Vector< typename Private::RCMul<NType1, NType2>::type >
			(
				lhs.y*rhs.z - lhs.z*rhs.y,
				lhs.z*rhs.x - lhs.x*rhs.z,
				lhs.x*rhs.y - lhs.y*rhs.x
			);
	}

	//XJ[Ƃ̏揜	
	
	template <typename NType1, typename NType2>
	const Vector< typename Private::RCMul<NType1, NType2>::type >
	operator*(const Vector<NType1>& v, const NType2& s){
		return Vector< typename Private::RCMul<NType1, NType2>::type >
			(v.x*s, v.y*s, v.z*s);
	}


	template <typename NType1, typename NType2>
	const Vector< typename Private::RCMul<NType1, NType2>::type >
	operator*(const NType1& s, const Vector<NType2>& v){
		return Vector< typename Private::RCMul<NType1, NType2>::type >
			(s*v.x, s*v.y, s*v.z);
	}


	template <typename NType1, typename  NType2>
	const Vector< typename Private::RCDev<NType1, NType2>::type >
	operator/(const Vector<NType1>& v, NType2 s){
		return Vector< typename Private::RCDev<NType1, NType2>::type >
			(v.x/s, v.y/s, v.z/s);
	}


	//xNg̎Od
	template <typename NType1, typename  NType2, typename  NType3>

	inline decltype(NType1(0) * NType2(0) * NType3(0) )
	TripleScalar(const Vector<NType1>& u,const Vector<NType2>& v,const Vector<NType3>& w){
		return 
			(u.x * (v.y*w.z-v.z*w.y)) +
			(u.y * (v.z*w.x-v.x*w.z)) +
			(u.z * (v.x*w.y-v.y*w.x));
	}







	//}Ci[֐
	//----------------------------------------------------------------------------------------------------------


	//xNg̕𓾂
	template <typename NType1, typename NType2>
	Vector<NType1> GetComponent(const Vector<NType1>& lhs, Vector<NType2> comp){
			comp.Normalize();
			auto d = (comp / ( NType2(1)*NType2(1) )) * lhs;
			return comp*d;
	}


	template <typename NType>
	Vector<NType> VAbs( const Vector<NType>& v){
		return Vector<NType>(abs(v.x), abs(v.y), abs(v.z) );
	}


	template <typename NType>
	//@xNg(3D͊OςłȂ)
	inline Vector<NType> GetNormalVector2D(const Vector<NType>& v){
		return Vector<NType>(-v.y,v.x,0.0);
	}


	//xNgԂ̃RTC
	template <typename NType>
	inline decltype( NType(0) / NType(0) )
	GetCos(Vector<NType>& lhs, Vector<NType>& rhs){
		NType mag1 = lhs.Length();
		NType mag2 = rhs.Length();
		if(abs(mag1)< NType(VEPSILON) || abs(mag2)< NType(VEPSILON) ) throw std::runtime_error("Vector is too small.");;
		return	(lhs.x*rhs.x + lhs.y*rhs.y + lhs.z*rhs.z) / (mag1 *mag2);
	}


	template <typename NType1, typename NType2>
	inline const Vector<NType1> VRotate2D(const Vector<NType1>& v, NType2 sin, NType2 cos){
		return Vector<NType1>(v.x*cos-v.y*sin, v.x*sin+v.y*cos, 0.0);
	}

	template <typename NType>
	inline decltype( NType(0) * NType(0) )
	VDistanceSq(const Vector<NType>& v1, const Vector<NType>& v2){
		return (v1-v2)*(v1-v2);
	}
	
	template <typename T>
	inline T VEuclideanDistance(const Vector<T>& v1, const Vector<T>& v2){
		return (v1-v2).Length();
	}
	
	template <typename T>
	inline T VManhattanDistance(const Vector<T>& v1, const Vector<T>& v2){
		T x = v1.x-v2.x;
		T y = v1.y-v2.y;
		T z = v1.z-v2.z;
		if(x<T(0))x=-x;
		if(y<T(0))y=-y;
		if(z<T(0))y=-z;
		return x+y+z;
	}
}

	

